/* The jQuery function declared here.
 * Simple one, we take an action and an input and we decide what to do.
 * 
 * @param {type} $
 * @returns {undefined}
 */
(function($) {
    /* serialize and unserialize functions converted from Locutus NodeJS Library into Client Javascript. */
    var PhpLikeSerializer = function() {
        /* Zero length codes of UTF8.
         *
         * @type Array
         */
        var zeroCodes = [
            338,
            339,
            352,
            353,
            376,
            402,
            8211,
            8212,
            8216,
            8217,
            8218,
            8220,
            8221,
            8222,
            8224,
            8225,
            8226,
            8230,
            8240,
            8364,
            8482
        ];

        /* Calculates the UTF8 size for a character.
         * The encoding wasn't doing the same of decoding with lengths so I had to change the first if.
         *
         * @param {type} str
         * @returns {Number}
         */
        var _utf8Size = function (str) {
            var size = 0;
            var i = 0;
            var l = str.length;
            var code = '';
            for (i = 0; i < l; i++) {
                code = str.charCodeAt(i);
                if (code < 0x0080 || code >= 0x00A0 && code <= 0x00FF || zeroCodes.indexOf(code) !== -1) {
                    size += 1;
                } else if (code < 0x0800) {
                    size += 2;
                } else {
                    size += 3;
                }
            }

            return size;
        };

        /* Removes the UTF8 overhead from a character.
         *
         * @param {type} chr
         * @returns {Number}
         */
        var utf8Overhead = function (chr) {
            var code = chr.charCodeAt(0);
            if (code < 0x0080 || code >= 0x00A0 && code <= 0x00FF || zeroCodes.indexOf(code) !== -1) {
                return 0;
            }
            if (code < 0x0800) {
                return 1;
            }
            return 2;
        };

        /* Returns the type of an input value (in Javascript).
         *
         * @param {type} inp
         * @returns {Array|PhpSerializer._getType.types|String}
         */
        var _getType = function (inp) {
            var match;
            var key;
            var cons;
            var types;
            var type = typeof inp;

            if (type === 'object' && !inp) {
                return 'null';
            }

            if (type === 'object') {
                if (!inp.constructor) {
                    return 'object';
                }

                cons = inp.constructor.toString();
                match = cons.match(/(\w+)\(/);
                if (match) {
                    cons = match[1].toLowerCase();
                }

                types = ['boolean', 'number', 'string', 'array'];
                for (key in types) {
                    if (cons === types[key]) {
                        type = types[key];
                        break;
                    }
                }
            }

            return type;
        };

        /* Read data from offset until a stop character.
         *
         * @param {type} data
         * @param {type} offset
         * @param {type} stopchr
         * @returns {Array}
         */
        var readUntil = function (data, offset, stopchr) {
            var i = 2;
            var buf = [];
            var chr = data.slice(offset, offset + 1);

            while (chr !== stopchr) {
                if ((i + offset) > data.length) {
                    console.log('Error: Invalid');
                    break;
                }

                buf.push(chr);
                chr = data.slice(offset + (i - 1), offset + i);
                i += 1;
            }

            return [buf.length, buf.join('')];
        };

        /* Read data from offset until length.
         *
         * @param {type} data
         * @param {type} offset
         * @param {type} length
         * @returns {Array}
         */
        var readChrs = function (data, offset, length) {
            var i, chr, buf;

            buf = [];
            for (i = 0; i < length; i++) {
                chr = data.slice(offset + (i - 1), offset + i);
                buf.push(chr);
                length -= utf8Overhead(chr);
            }

            return [buf.length, buf.join('')];
        };

        /* Serializes (like PHP) a mixed value.
         * This method doesn't save passed object methods, but if you want to save an attribute which is a function,
         * you should add 'AttributeFn' at the end of the attribute name in your object.
         * Example: myFun => myFunAttributeFn
         *
         * @param {type} mixedValue
         * @returns {String}
         */
        this.serialize = function(mixedValue) {
            var val, key, okey;
            var ktype = '';
            var vals = '';
            var count = 0;

            var type = _getType(mixedValue);

            switch (type) {
                case 'function':
                    //Removes new liners
                    var stringVersion = mixedValue.toString().replace(/[\r\n]/g, '');
                    //Escape double quotes for avoid code to break
                    stringVersion = stringVersion.replace(/"/g, '\\"');

                    val = 'f:' + _utf8Size(stringVersion) + ':"' + stringVersion + '"';
                    break;
                case 'boolean':
                    val = 'b:' + (mixedValue ? '1' : '0');
                    break;
                case 'number':
                    val = (Math.round(mixedValue) === mixedValue ? 'i' : 'd') + ':' + mixedValue;
                    break;
                case 'string':
                    val = 's:' + _utf8Size(mixedValue) + ':"' + mixedValue + '"';
                    break;
                case 'array':
                case 'object':
                    val = 'a';

                    for (key in mixedValue) {
                        if (mixedValue.hasOwnProperty(key)) {
                            ktype = _getType(mixedValue[key]);

                            if (ktype === 'function') {
                                /* Only saves a function if it's an attribute and not an object
                                 * method. */
                                if(!key.match(/AttributeFn/i))
                                    continue;
                            }

                            okey = (key.match(/^[0-9]+$/) ? parseInt(key, 10) : key);
                            vals += this.serialize(okey) + this.serialize(mixedValue[key]);
                            count++;
                        }
                    }
                    val += ':' + count + ':{' + vals + '}';
                    break;
                case 'undefined':
                default:
                    val = 'N';
                    break;
            }

            if (type !== 'object' && type !== 'array') {
                val += ';';
            }

            return val;
        };

        /* Unserializes (like PHP) a serialized value.
         * This is wrapper for _unserialize private method, which is recursive.
         *
         * @param {type} data
         * @returns {unresolved}
         */
        this.unserialize = function(data) {
            return _unserialize((data + ''), 0)[2];
        };

        /* The real unserialize function, recursive.
         *
         * @param {type} data
         * @param {type} offset
         * @returns {Array}
         */
        var _unserialize = function (data, offset) {
            var dtype;
            var dataoffset;
            var keyandchrs;
            var keys;
            var contig;
            var length;
            var array;
            var readdata;
            var readData;
            var ccount;
            var stringlength;
            var i;
            var key;
            var kprops;
            var kchrs;
            var vprops;
            var vchrs;
            var value;
            var chrs = 0;

            var typeconvert = function (x) {
                return x;
            };

            if (!offset) {
                offset = 0;
            }

            dtype = (data.slice(offset, offset + 1)).toLowerCase();
            dataoffset = offset + 2;

            switch (dtype) {
                case 'i':
                    typeconvert = function (x) {
                        return parseInt(x, 10);
                    };

                    readData = readUntil(data, dataoffset, ';');
                    chrs = readData[0];
                    readdata = readData[1];
                    dataoffset += chrs + 1;
                    break;
                case 'b':
                    typeconvert = function (x) {
                        return parseInt(x, 10) !== 0;
                    };

                    readData = readUntil(data, dataoffset, ';');
                    chrs = readData[0];
                    readdata = readData[1];
                    dataoffset += chrs + 1;
                    break;
                case 'd':
                    typeconvert = function (x) {
                        return parseFloat(x);
                    };

                    readData = readUntil(data, dataoffset, ';');
                    chrs = readData[0];
                    readdata = readData[1];
                    dataoffset += chrs + 1;
                    break;
                case 'n':
                    readdata = null;
                    break;
                case 's':
                    ccount = readUntil(data, dataoffset, ':');
                    chrs = ccount[0];
                    stringlength = ccount[1];
                    dataoffset += chrs + 2;

                    readData = readChrs(data, dataoffset + 1, parseInt(stringlength, 10));
                    chrs = readData[0];
                    readdata = readData[1];
                    dataoffset += chrs + 2;
                    if (chrs !== parseInt(stringlength, 10) && chrs !== readdata.length) {
                        console.log('SyntaxError: String length mismatch');
                        break;
                    }
                    break;
                case 'a':
                    readdata = {};

                    keyandchrs = readUntil(data, dataoffset, ':');
                    chrs = keyandchrs[0];
                    keys = keyandchrs[1];
                    dataoffset += chrs + 2;

                    length = parseInt(keys, 10);
                    contig = true;

                    for (i = 0; i < length; i++) {
                        kprops = _unserialize(data, dataoffset);
                        kchrs = kprops[1];
                        key = kprops[2];
                        dataoffset += kchrs;

                        vprops = _unserialize(data, dataoffset);
                        vchrs = vprops[1];
                        value = vprops[2];
                        dataoffset += vchrs;

                        if (key !== i) {
                            contig = false;
                        }

                        readdata[key] = value;
                    }

                    if (contig) {
                        array = new Array(length);
                        for (i = 0; i < length; i++) {
                            array[i] = readdata[i];
                        }

                        readdata = array;
                    }

                    dataoffset += 1;
                    break;
                case 'f':
                    typeconvert = function (x) {
                        var newFunction;
                        eval('newFunction = ' + x);

                        return newFunction;
                    };

                    ccount = readUntil(data, dataoffset, ':');
                    chrs = ccount[0];
                    stringlength = ccount[1];
                    dataoffset += chrs + 2;

                    readData = readChrs(data, dataoffset + 1, parseInt(stringlength, 10));
                    chrs = readData[0];
                    readdata = readData[1];
                    dataoffset += chrs + 2;
                    if (chrs !== parseInt(stringlength, 10) && chrs !== readdata.length) {
                        console.log('SyntaxError: String length mismatch');
                        break;
                    }

                    /* Unescape the escaped double quotes */
                    readdata = readdata.replace(/\\"/g, '\"');

                    break;
                default:
                    console.log('SyntaxError: Unknown / Unhandled data type(s): ' + dtype);
                    break;
            }

            return [dtype, dataoffset - offset, typeconvert(readdata)];
        };
    };

    var phpSerializer = new PhpLikeSerializer();

    $.phpSerialize = function(input) {
        return phpSerializer.serialize(input);
    };

    $.phpUnserialize = function(input) {
        return phpSerializer.unserialize(input);
    };
}(jQuery));